// matlab.e

/* Matlab mode for Epsilon
   Version 1.0
   1/21/98

   This epsilon extension includes auto-indenting, syntax coloring,
   and tagging for The MathWorks Inc. MATLAB(R) source code, version
   5.2 and earlier.

   Matlab mode is automatically invoked on files with extension .m.
   Several variables alter the behavior of matlab mode. They all
   begin with the prefix m-, e.g. m-indent, m-comment-flow, etc. A
   list can be obtained from matlab.h or by using epsilon's variable
   completion on m-. Matlab mode defines a new keytable called m_tab.

   Matlab Auto-indenting
    - The indentation level is controlled by the buffer specific
      variable m-indent. If this variable is zero the current buffer
      tab-size will be used.
    - Indentation occurs everytime the <enter> key is hit, which also
      opens a new line. To indent only, without a new line, use the
      command m-indent-only which is bound to ALT-<enter>
    - When an end, else, otherwise, or case is entered matlab mode
      will reindent the current line. This can be turned off by
      setting m_auto_reindent to zero.
    - Matlab mode includes delimiter matching like epsilon's c-mode.
      It also features end and else statement matching. When the
      keyword "end" or "else" is entered matlab mode will show the
      matching if, for, while, or switch statement. Of course else
      will only match an if. To invoke a match on the most recent
      previous end, else or elseif, use the command m-find-delimiter
      which is bound to Alt-]. If an unmatched  end or else statement
      is entered the bell will sound. This may be turned off by setting
      m_want_bell to zero. The end matching feature may be turned off
      by setting m-matchdelim to zero; 
    - In-line comments (comments following matlab statements) will be
      indented to the value of m_comment_column if it is non-zero. If
      it is zero the indentation will be to the value of epsilon's
      variable comment-column. To turn off identing of in-line
      comments set m-inline-indent to zero.
    - When entering a comment on a line by itself (full-line) an
      <enter> key will start a new comment on the next line. This may
      be turned off by setting m-comment-flow = 0. If the comment is
      an in-line comment the <enter> key will not generate a new line
      but the indentation will be for a new statement. If m-auto-fill
      is non-zero both full-line and in-line comments will auto-fill
      at the buffer's current fill column. When hitting <enter> on an
      empty comment line, the line with the empty comment will be
      deleted. To prevent deleting of empty comment lines set
      m-comment-delete-empty = 0. 

   Matlab Syntax coloring
    - As with other epsilon coloring modes, to turn it off set
      want-code-coloring to zero.
    - Matlab coloring does not introduce any new color classes. It
      uses the C color classes in the obvious ways. This allows us to
      piggyback onto existing color schemes. To define a new matlab
      coloring scheme or edit an old scheme you must set the
      appropriate C color classes. They are: 
             c-comment     -- % this is a comment
             c-string      -- 'this is a string'
             c-punctuation -- =,*,-,^,.^,.*,[],(),{},;,:,etc
             c-number      -- 1.5e-10, 3.14, ...
             c-keyword     -- for, if, else, end, global, ...
             c-function    -- inv(.), etc
             c-identifier  -- variables
      It is generally best to assign function and identifier classes
      the same color since one cannot differentiate between an
      indexed array and a function.
      
   Matlab Tagging
    - Matlab tagging was developed primarily to support long function
      switchyards which are used to write GUIs. It also tags function
      names, including subfunctions. The original motivation was to
      tag constructs such as
         if strcmp(action,'tag1')
           a=1;  ...
         elseif strcmp(action,'tag2')
           a=2;  ...
         elseif findstr(action,'tag3')
           a=3;  ...
           .
           .
         end
      In Matlab 5 this became
         switch action
         case 'tag1'
           a=1; ...
         case 'tag2'
           a=2; ...
           .
           .
         end
      Both of these are supported. Currently elseif and case are only
      tagged if they begin in column 0.
    - To tag matlab files (*.m files) run the epsilon command
      tag-files, bound to Ctrl-X Alt-., and provide a file or a
      wildcard specification of a set of .m files. The epsilon
      tag commands such as goto-tag, pluck-tag, etc can now be used.
    - If m-tag-function is non-zero, lines which begin with
         function [x,y,...] = tagident(a,b,...)
      will be tagged with the tag "tagident"
    - If m-tag-case is non-zero, lines which begin with
         case 'tagident'
      will be tagged with the tag "tagident". This version does not
      tag cell array expressions, i.e., case {'hello','goodbye'}
    - If m-tag-elseif is non-zero, lines which begin with
         elseif anyfunc(a,b,..,'tagident',...)
      will be tagged with the tag "tagident". The last string which
      encloses a single identifer on that line will be the tag.
    - The code between two tagged lines is called a section (this
      will depend on the settings of m-tag-function, m-tag-case, and 
      m-tag-elseif). The command narrow-to-section on Alt-n will mark
      a section as a region and invoke narrow-to-region. Alt-n is a
      toggle and will widen the buffer if currently narrowing.
    - To move from tagged line to tagged line in the current buffer
      use find_next_section on Alt-. and find_previous_section on
      Alt-,.

The code in this extension relies heavily on epsilon's c.e and colcode.e
      
Please send any bugs or comments on this matlab mode to:

Terry Brennan
the Optical Sciences Company
tbrennan@tosc.com

*/

/* note: there is an eel bug in find_group when doing a reverse search
         the workaround - replace
           find_group(n,1) with MIN(find_group(n,1),find_group(n,0))
           find_group(n,0) with MAX(find_group(n,1),find_group(n,0))

   note 2: When doing narrowing with coloring some versions of
           epsilon may consume all the cpu time. This was fixed with
           a patch from Lugaru, cpu.e. This may be unnecessary in an
           epsilon version > 9.0.
*/

#include "eel.h"
#include "c.h"
#include "matlab.h"
#include "colcode.h"

// Automatic indentation for m code (Matlab)

m_indenter() // matlab indents
{
  indent_to_column(m_compute_indent());
}

command matlab_mode()
{
  mode_keys = m_tab;
  // It seems that it a better approach would be for show_matching_delimiter
  // to interogate Matchdelim. With the currrent approach you cannot turn
  // Matchdelim on or off without re-entering matlab mode.
  m_tab[']'] = Matchdelim ? (short) show_matching_delimiter : 0;
  m_tab[')'] = Matchdelim ? (short) show_matching_delimiter : 0;
  m_tab['}'] = Matchdelim ? (short) show_matching_delimiter : 0;
  // remove binding of 'f' left over from previous version
  m_tab['f'] = 0;
  indenter = m_indenter;
  auto_indent = 1;
  major_mode = _m_mode_name;
  strcpy(comment_start, "[^\\]%%[ \t\f]*"); // check this
  strcpy(comment_pattern, "[^\\]%%.*$");
  strcpy(comment_begin, "% ");
  strcpy(comment_end, "");
  recolor_range = color_m_range;  // set up coloring rules
  recolor_from_here = color_m_from_here;
  if (want_code_coloring)   // maybe turn on coloring
    when_setting_want_code_coloring();
  if (m_mode_mouse_to_tag)
    mouse_goes_to_tag = 1;
  make_mode();
}

suffix_m()
{
  matlab_mode();
}

jmp_buf m_indent_giveup;  /* come here when we hit narrowed edge */

/*  Look at the last two lines and return the correct indentation
    for the current line, assuming M source (or similar).
    Also reindents previous line as necessary.
    At start, we must be at the end of a line's indentation.
*/

m_compute_indent()
{
  int ind = 0;  // indentation to use 
  int pstart;   // start position of prev line
  int trailing_comment;
  int flowing_comment = 0;
  spot pcurrent = alloc_spot();

  if (setjmp(&m_indent_giveup)) // if we hit end & gave up
    return ind;                 // use best guess so far 
  
  to_begin_line();
  mind_space();     // skip whitespace
  trailing_comment = (is_trailing_comment() &&
                      no_blank_lines(*pcurrent) &&
                      m_comment_column);

  to_m_statement_start(); // first significant character
  pstart = point;

  if (is_m_else()) {   // re-indent end, else, and elseif on 
    ind = get_matching_start_column(); // previous line
    to_m_statement_start();
    indent_to_column(ind);
  }
  else if (is_m_case(pstart)) { // case followed by a case or otherwise
                                // should be aligned
    ind = get_previous_case_column();
    to_m_statement_start();
    indent_to_column(ind);
  }
  else if (character(pstart)=='%' && m_comment_flow &&
                        no_blank_lines(*pcurrent)) {
    if (flow_comment(pstart,*pcurrent))
      *pcurrent = point; // adjust the return point
    ind = current_column();
    point = pstart;
  }
  else if (trailing_comment && (key != '\n')) {
    ind = get_m_comment_column();
    if (parse_string(1,".+(%%)")) {
      point = find_group(1,1);
       // double duty for flow_comment not good
      if (flowing_comment = flow_comment(point,*pcurrent)) {
        *pcurrent = point;
        ind = current_column();
        point = pstart;
      }
    }
  }
  else
    ind = current_column();

  pstart = point;

  // indentation for current line
  if (first_continuation() || first_matrix())
    ind = current_column();
  else {
    point = pstart;
    if (last_continuation())
      ind = current_column();
    else {
      point = pstart;
      if (last_matrix()) {
        if (get_assignment_start())
          ind = current_column();
      }
      else {
        if (m_statement_start()) // previous line starts a group
          if (!m_statement_ended() && !flowing_comment) // not if it is ended on same line
            ind += get_m_indent();
      }
    }
  }
    
  // reindent trailing comments on previous line
  if (m_comment_column) 
    indent_m_comment();
  point = *pcurrent;
  free_spot(pcurrent);
  return ind;
}

command m_indent_alone() on m_tab[ALT(CTRL('M'))]
{
  save_spot point;
  to_end_line();
  point++;
  m_compute_indent();
}
      
indent_m_comment() // looks for next comment on line and indents it
{
  int ok_to_indent;
  if (parse_string(1,"[^\n%%]+(%%)")) {
    point = find_group(1,1);
     /* this is a simple but not always correct test that the current
        % is not in a format statement
        The second condition says don't indent when multiple %'s on one line
     */
    ok_to_indent = (!is_in_string() && !parse_string(1,"%%.*%%"));
    if (ok_to_indent) {
      indent_to_column(get_m_comment_column());
      // Make sure there is whitespace between end of statement and %
      if (!index(" \t", character(point - 1)))
        indent_to_column(current_column()+1);
    }
  }
}

// assume first line of multi-line statement has an equal sign
int first_continuation()
{
  if (parse_string(1,".*=[ \t]*([^ ].*%.%.%.)")) {
    point = find_group(1,1);
    point += parse_string(1,"([a-zA-Z0-9_]*[ ]*%(|%[)");
    return 1;
  }
  return 0;
}

// Opening line of a matrix assignment, which is not completed and
// not continued with ..., i.e. A = [1 2 3
// We assume here that we are NOT on the first line of a continuation
int first_matrix()
{
  if (parse_string(1,".*=[ \t]*(%[)")) {
    point = find_group(1,1)+1;
    if (parse_string(1,"[ \t]*(.)"))
      point = find_group(1,1);
    return 1;
  }
  return 0;
}

// previous line finished a matrix assignment
int last_matrix()
{
  save_var point;
  // move past [..] pairs
  while (parse_string(1,".*%[[^%]]*%]"))
    point = matchend;
  return parse_string(1,".*%]");
}

// previous line finished a multi-line set
int last_continuation()
{
  int doindent = 0;
  if (!parse_string(1,".*%.%.%.")) { // this line is not continued
    while (m_is_contin()) {
      point--;
      to_begin_line();
      doindent = 1;
    }
    if (doindent) 
      to_indentation();
  }
  return doindent;
}
     
mind_space()    /* move back over any whitespace or comments */
{     /* give up if we hit start */
  if (mind_try_space())
    return;  // haven't really correctly implemented this longjmp yet
//    longjmp(&m_indent_giveup, 1);
}

mind_try_space()  // move back over whitespace, comments
{                 // lines, and return 0, or 1 if we hit start
  for (;;) {
    if (index(" \t\n\f", character(point - 1)))
      point--;
    else
      return point == narrow_start;
  }
}

int is_m_else() // is there an else, elseif, or end keyword at point
{
  if (character(point) != 'e')
    return 0;
  return parse_string(1, "(else|elseif|end)[^a-zA-Z0-9_]");
}

int is_m_case(p) // assume at start of possible case or otherwise
int p;
{
  save_var point = p;
  return parse_string(1, "(case|otherwise)[^a-zA-Z0-9_]");
}

int get_previous_case_column()
{
  int col, col0 = current_column();
  save_var point;
  while (re_search(-1,"(end|case|switch)[^a-zA-Z0-9_]")) {
    if (!parse_string(-1,"[a-zA-Z_]"))
      if (!is_in_m_comment() && !is_in_string() &&!is_end_index()) { // it is a statement
        if (character(point) == 'e') {// found an end move past its start
          if (!get_matching_start(0))
            break;
        } else {
          col = current_column();
          if (character(point) == 's') // found switch so indent
            col += get_m_indent();
          return col;
        }
    }
  }
  return col0;
}

/* Checks if the previous statement is an end */
int m_statement_ended()
{
  int psave = point;
  save_var point;
  if (parse_string(1,"(.+)%%")) // go to beginning of trailing comment
    point = find_group(1,0);
  else {
    point = psave;
    nl_forward();
    point--;
  }
  return parse_string(-1,"[^a-zA-Z0-9_]end[ \t\f,;]*");
}
   
to_m_statement_start()
{
  to_begin_line();
  to_indentation();
}
/*
Tell if the last line begins a statement (rather than a function
declaration or the continuation of a previous statement).
Assumes point is at the start of the last line.
*/
m_statement_start()
{
  return parse_string(1, "[ \t\f]*(if|else|elseif|while|for|switch|"
                         "case|otherwise)[^a-zA-Z0-9]");
}

int get_assignment_start()
{
  return re_search(-1,"[a-zA-Z][a-zA-Z0-9_]*[ \t]*=");
}

int get_matching_start_column()
{
  save_var point;
  get_matching_start(character(point+1) == 'l'); // match else or end
  return current_column(); // if not found it will return current column
}

/* match else or end to their if,while,for,or switch
   assumes point is just before end or else */
int get_matching_start(iselse)
int iselse;
{
  int extra_ends = 0;
  // Also must check for embedded ends
  while (re_search(-1,"(end|else|if|while|for|switch)[^a-zA-Z0-9_]")) {
    if (!parse_string(-1,"[a-zA-Z0-9_]")) { // Not in first re_search to handle the
                                            // case of for at beginning of file
      if ((!is_in_m_comment() && !is_end_index()) && !is_in_string()) {
        if (character(point) == 'e') { // end or else
          if (character(point+1) == 'n') 
            extra_ends++;
          else // if an else was entered and the first find was an else there is a error
            if (iselse && !extra_ends) 
              return 0;
        }
        else if (--extra_ends<0) // match found: is it correct
          return !(iselse && character(point)!='i'); // an else must be matched with an if
      }
    }
  }
  return 0;
}

/* is end used as an index rather than a group end, could be an array index
   or a cell array index */
int is_end_index()
{
  int isndx = 0, psave = point, start;
  save_var point;
  if (!parse_string(1,"end"))
    return 0;
  to_statement_start();
  start = point;
  point = psave;
  while (re_search(-1,"(%(|%))")) { // ordinary array
    if (point < start) // no need to search beyond start of statement
      break;
    if (character(point) == ')')
      isndx++;
    else if (--isndx<0)
      return 1;
  }
  point = psave;
  while (re_search(-1,"({|})")) { // cell array
    if (point < start)
      break;
    if (character(point) == '}')
      isndx++;
    else if (--isndx<0)
      return 1;
  }
  return 0;
}

// Attempts to match end statement
// Removed binding to 'f', This worked but caused elseif to find the if twice
// once for the 'e' and once for the 'f'
command m_show_matching_delimiter() on m_tab['d'], m_tab['e'] //, m_tab['f']
{
  normal_character();
  if (!is_in_m_comment() && !is_in_string()) {
    if (m_auto_reindent && parse_string(-1,"end|else|case|otherwise"))
      m_indent_alone();
    if (m_matchdelim)
      m_find_delimiter();
  }
}

command m_find_delimiter() on m_tab[ALT(']')]
// Alt-] will match elseif but we have unbound it from m-show-matching-delimiter
{
  char buf[7];
  int iselse;
  save_var point;
  
  if (parse_string(-1,"[^a-zA-Z0-9_]+(end|else|elseif)[^a-zA-Z0-9_]*")) {
    grab(point=MIN(find_group(1,0),find_group(1,1)),
               MAX(find_group(1,0),find_group(1,1)),buf);
    if (is_end_index())
      return;
    else
      iselse = (character(point+1) == 'l'); // find delimiter for else
    say("");
    iter=0;
    if (in_macro())
      return;
    if (get_matching_start(iselse))
      show_line();
    else {
      maybe_ding(m_want_bell);
      say("Unmatched %s statement",buf);
    }
  }
}

command m_comment_handling() on m_tab[' ']
{
  int broke = 0;
  spot old;
  if (!is_in_m_comment()) {
    maybe_break_line();
    return;
  } else if (is_trailing_comment() && !m_inline_ident) {
    normal_character();
    return;
  }
  if (m_auto_fill) {
    old = alloc_spot();
    if (m_comment_column && is_trailing_comment()) {
      if (character(point-1)=='%') {
          to_begin_line();
          indent_m_comment();
          point = *old;
      } else
        broke = m_maybe_break_line();
    }
    else if (is_in_m_comment())
      broke = m_maybe_break_line();
    free_spot(old);
    if (broke && ((character(point-1)==' ') || (character(point-1)=='\t')) )
      return;
  }
  normal_character();
}

// Is point inside of trailing comment
is_trailing_comment()
{
  return parse_string(-1,"[^ \t\n%%]+[ \t]*%%*");
//  return parse_string(-1,"[^ \t\n%%]+[ \t]*%%");
}

// this was swiped from format.e and modified.
m_maybe_break_line()
{
  int h, found = 1;
  spot old;

  if (m_auto_fill && is_in_m_comment()) {
    h = current_column() + 1;
    if (h == margin_right) {
        break_line_here();
        return 1;
    } else if (h > margin_right) {
      old = alloc_spot();
      found = pick_break(margin_right);
      if (found) {
        break_line_here();
        key = ' ';
        point = *old;
      } else
        break_line_here();
      free_spot(old);
      return 1;
    }
    return 0;
  }
  return 0;
}

no_blank_lines(there) /* checks for blank lines between point and there */
int there;
{
  save_var point;
  if (re_search(1,"\n[ \t]*\n"))
    return point > there;
  return 1;
}
  
int flow_comment(here,there)
int here, there;
{
  int col0, col;
  point = here;
  col0 = current_column();
  if (parse_string(1,"%%[ \t]*")) // single % followed by whitespace
    point = find_group(0,0);
  // end of line => empty comment, delete it
  if ((character(point) == '\n') && m_comment_delete_empty) { 
    delete(find_group(0,1),point);
    return 0;
  } else {
    col = current_column();
    point = there;
    indent_to_column(col0);
    insert('%');
    indent_to_column(col);
    return 1;
  }
}
  
// This doesn't deal with format statements
int is_in_m_comment()
{
  return parse_string(-1,"%%.*");
}

int is_in_string()
{
  int t = point;
  save_var point;
  to_begin_line();
  while (parse_string(1,"<FE>.*\'"))  {
    point = matchend;
    if (point >= t)
      return 0;
    if ( definitely_not_transpose(point-1) ) { // we are at left quote
      find_right_quote();
      if (point>t)
        return 1;
    }
  }
  return 0;
}

get_m_indent() // indent this many spaces for each level
{              // if m_indent is 0, use tab_size instead
  return (m_indent > 0) ? m_indent : tab_size;
}

get_m_comment_column()
{
  return (m_comment_column > 0) ? m_comment_column : comment_column;
}

//
// Matlab Coloring we make use of many c coloring functions
//

color_m_from_here(safe) // Move backward to the nearest line guaranteed
{     // to start outside any colored region, return pos.
  to_begin_line(); // begin line should be safe for m
  while (m_is_contin()) {
    point--;
    to_begin_line();
  }
  if (point<safe)
    point = safe;
  return point;
}

/* is the current line a continuation of the previous
   it is assumed that comments can be at the end of continuation
   lines but not on a line alone although they can in certain
   cases */
m_is_contin()
{
  save_var point;
  if (nl_reverse()) {
    point -= parse_string(-1,"%%.*"); // move to start of possible comment
    return parse_string(-1,"%.%.%.[ \t]*");
  }
  return 0;
}
  
color_m_range(from, to) // recolor just this section
{     // last colored region may go past to
  int t = -1, talk;
  char pat[200];

  if (from >= to)
    return to;
  /* the following line was suggested by Doerfler, to prevent hogging of cpu
     during narrowing. It may not be needed in the next interim version (It
     also doesn't seem to help) Problem was solved with cpu.e from Lugaru */
//  save_var narrow_start = 0, narrow_end = 0;  save_var point, matchstart, matchend;
  c_init_color(from, to);
  point = from;
  talk = (to - from > 2000);  // show status during long delays
  save_var case_fold = 0;
  strcpy(pat, "%%|\'");   // start of comment or quote
  if (!minimal_coloring)
    strcat(pat, "|[A-Za-z][A-Za-z0-9_]*" // identifier
           "|-?(%.)([0-9]|([Ee]|[Dd])[0-9-+])"       // numbers
           "|-?([0-9])([0-9]|([Ee]|[Dd])[0-9-+])*"); // more numbers
  while (point < to) {
    if (!re_search(1, pat)) {
      t = size();
      break;
    }
    t = matchstart;
    switch (character(point - 1)) {   // check last char
    case '%':   // found comment
      while (nl_forward() && character(point) == '%')
        ;   // Continued // comment.
      set_character_color(t, point, color_class c_comment);
      break;
    case '\'':  // found a string literal or a transpose
      if ( definitely_not_transpose(t) ) { // we are at a left quote
        find_right_quote();
        set_character_color(t, point, color_class c_string);
      } else    // we are at a transpose operator
        set_character_color(t, point, color_class c_punctuation);
        break;
    default:    // found identifier, kywd, or number
      set_character_color(t, point, m_keyword_color(t));
      break;
    }
    if (talk)
      note("Coloring M program: %d%% complete...",
           (point - from) * 100 / (to - from));
  }
  c_init_color(to, t);
  if (talk)
    note("");
  return point;
}

int definitely_not_transpose(t)
int t;
{
  if (index(" =,;%[%({\n\t", character(t-1)))
    return 1;
  return 0;
}

find_right_quote() // This assumes you are in a string
{
  int rightfound = 0;
  while (!rightfound) {
    rightfound = 1;
    if (re_search(1,"[.]*\'"))
      rightfound = skip_quotes();
  }
}

int skip_quotes()
{
  int i = 1;
  while(character(point)=='\'') {
    point++;
    i++;
  }
  return i % 2;
}

m_keyword_color(from) // return color for "identifier" from here to point
{     // (something with alpha or digits)
  char buf[500];

  if (point - from > sizeof(buf) - 10)
    save_var point = from + sizeof(buf) - 10;
  buf[0] = '|';   // get identifier, between | chars
  grab(from, point, buf + 1);
  if (index("0123456789-.", buf[1]))
    return c_number_color(buf + 1);
  strcpy(buf + point - from + 1, "|");
  if (is_m_keyword(buf) && character(from-1) != '.')
    return color_class c_keyword;
  if (color_class c_function != color_class c_identifier
      && paren_follows())
    return color_class c_function;
  return color_class c_identifier;
}

is_m_keyword(p)   // is text in p (must be surrounded by |'s) a keyword?
char *p;
{
  if (strstr("|if|else|elseif|end|while|for|return|break|switch|"
       "case|otherwise|global|function|keyboard|",p))
    return 1;
  return 0;
}

//
// Matlab Code for tags and jumping
//

tag_suffix_m()
{
  char func[TAGLEN];
  int start, ofold = case_fold;

  case_fold = 0;
  while (find_elseif_section(1)) {
    grab(start = find_group(1, 1), find_group(1, 0), func);
    add_tag(func, start);
  }
  point = 0;
  while (find_case_section(1)) {
    grab(start = find_group(1, 1), find_group(1, 0), func);
    add_tag(func, start);
  }
  point = 0;
  while (find_function_section(1)) {
    grab(start = find_group(1, 1), find_group(1, 0), func);
    add_tag(func, start);
  }
  case_fold = ofold;
}

command find_next_section() on m_tab[ALT('.')]
{
  int temp = point;
  nl_forward();
  if (!find_section(1)) {
    point = temp;
    say("No more sections.");
  }
}

command find_previous_section() on m_tab[ALT(',')]
{
  int temp = point;
  nl_reverse();
  if (!find_section(-1)) {
    point = temp;
    say("No more sections.");
  }
}

int find_section(dir)
int dir;
{
  int pe, pstart=point;
  char rtn=0;
  if (dir>0)
    pe = size();
  else
    pe = 0;
  if (find_elseif_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = matchstart;
    else
      pe = point;
  } else
  point = pstart;
  if (find_case_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = MIN(matchstart,pe);
    else
      pe = MAX(point,pe);
  }
  point = pstart;
  if (find_function_section(dir)) {
    rtn = 1;
    if (dir>0)
      pe = MIN(matchstart,pe);
    else
      pe = MAX(point,pe);
  }
  if (rtn)
    point = pe;
  else
    point = pstart;
  return rtn;
}

int find_elseif_section(dir)
int dir;
{
  if (!m_tag_elseif)
    return 0;
  return re_search(dir,"^elseif.+'([A-Za-z0-9_ ]+)'");
}

int find_case_section(dir)
int dir;
{
  if (!m_tag_case)
    return 0;
  return re_search(dir,"^case[ ]+'([A-Za-z0-9_ ]+)'");
}

int find_function_section(dir)
int dir;
{
  if (!m_tag_function)
    return 0;
  return re_search(dir,"^function[^=\n%(]+[ =\t]*([A-Za-z0-9_]+)[ \t]*%(?");
}
  
command narrow_to_section() on m_tab[ALT('n')]
{
  if ( (narrow_start>0) || (narrow_end>0) ) 
    widen_buffer();
  else {
    save_var point;
    point++;
    if (!find_section(1))
      point = size();
    set_mark();
    if (find_section(-1)) {
      to_begin_line();
      narrow_to_region();
    }
  }
}

command matlab_show_version()
{
  say("Matlab Mode %s", matlab_version);
}
